/*
 * tkv4lng.c --
 *
 *	Tk V4L(2) next generation widget
 */

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <fcntl.h>
#include <X11/X.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/extensions/XShm.h>
#ifdef XV_SUPPORT
#include <X11/extensions/Xvlib.h>
#endif
#include <unistd.h>
#include <string.h>
#include <stdlib.h>
/* This define allows this module to be compiled with Tk8.5 */
#define USE_PANIC_ON_PHOTO_ALLOC_FAILURE
#include <tk.h>
#if (TK_MAJOR_VERSION < 8)
#error "need Tk 8.0 or higher"
#endif
#if (TK_MAJOR_VERSION == 8) && (TK_MINOR_VERSION == 0)
#define TK80 1
#endif

#ifndef CONST
#define CONST
#endif

#include "grab-ng.h"

#define SHM_WIDTH  1024
#define SHM_HEIGHT 1024
#define SHM_SIZE (SHM_WIDTH * SHM_HEIGHT * 4)

/*
 * Flag bits for V4LNG:
 *
 * REDRAW_PENDING:		Non-zero means a DoWhenIdle handler
 *				has already been queued to redraw
 *				this window.
 * COMPL_PENDING:
 */

#define REDRAW_PENDING		1
#define REDRAW_XEVENTS		2
#define REDRAW_NOCLIP		2
#define COMPL_PENDING		8
#define REDRAW_AGAIN		16
#define GOT_FOCUS		32
#define REQUEST_REDRAW		64
#define UPDATE_XVIEW		128
#define UPDATE_YVIEW		256
#define DEV_CAPTURE		512
#define RECORDING		1024

#ifdef XV_SUPPORT
/*
 * Make FOURCC in network byte order.
 */

#define FOURCC(x0,x1,x2,x3)	htonl(((x0)<<24)|((x1)<<16)|((x2)<<8)|(x3))
#endif

/*
 * A data structure of the following type is kept for each
 * V4LNG that currently exists for this process:
 */

typedef struct {
    Tk_Window tkwin;		/* Window that embodies the V4LNG.  NULL
				 * means that the window has been destroyed
				 * but the data structures haven't yet been
				 * cleaned up. */
    Display *display;		/* Display containing widget.  Used, among
				 * other things, so that resources can be
				 * freed even after tkwin has gone away. */
    Tcl_Interp *interp;		/* Interpreter associated with widget. Used
				 * to delete widget command. */
    Tcl_Command widgetCmd;	/* Token for V4LNG's widget command. */
    Tk_Cursor cursor;		/* Current cursor for window, or None. */
    char *takeFocus;		/* Value of -takefocus option;  not used in
				 * the C code, but used by keyboard traversal
				 * scripts.  Malloc'ed, but may be NULL. */
    Tk_3DBorder border;		/* Structure used to draw 3-D border.
				 * NULL means no border. */
    int borderWidth;		/* Width of 3-D border (if any). */
    int relief;			/* 3-D effect: TK_RELIEF_RAISED etc. */
    int width;			/* Width to request for window.  <= 0 means
				 * don't request any size. */
    int height;			/* Height to request for window.  <= 0 means
				 * don't request any size. */
    int highlightWidth;		/* Width in pixels of highlight to draw
				 * around widget when it has the focus.
				 * 0 means don't draw a highlight. */
    XColor *highlightBgColorPtr;/* Color for drawing traversal highlight
				 * area when highlight is off. */
    XColor *highlightColorPtr;	/* Color for drawing traversal highlight. */
    char *xScrollCmd;		/* Command prefix for communicating with
				 * horizontal scrollbar.  NULL means no command
				 * to issue.  Malloc'ed. */
    char *yScrollCmd;		/* Command prefix for communicating with
				 * vertical scrollbar.  NULL means no command
				 * to issue.  Malloc'ed. */
    GC iGC;			/* GC for drawing XImage. */
    Region region;		/* Region to optimize expose/XPutImage. */
    int noxv;			/* When true, don't use XV extension. */
#ifdef XV_SUPPORT
    int hasxv;			/* True when XV extension available. */
    int usexv;			/* True when using Xv*PutImage(). */
    int xvCompletion;		/* XV event base. */
    int xvPort;			/* Port for Xv*PutImage(). */
    int xvMaxWidth, xvMaxHeight;/* Max. supported image size. */
    int xvNfmts;		/* Number of supported image formats. */
    XvImageFormatValues *xvFmts;	/* Supported image formats. */
    XvImage *xvImg;		/* XvImage for Xv*PutImage(). */
#endif
    int islocal;		/* True for local X display connection. */
    XShmSegmentInfo shmInfo;	/* XShm IPC descriptor. */
    int Completion;		/* XShm completion event number. */
    XImage *img;		/* XImage for V4LNG frame buffer. */
    int xoff, yoff;		/* X/Y view offsets for scrolling. */
    int vxoff, vyoff;		/* Computed view offsets for drawing. */
    int vwidth, vheight;	/* Requested video height. */
    int viw, vih;		/* Ditto, image dimension. */
    int scanMarkX, scanMarkY;	/* For widget "scan" minor command. */
    int flags;			/* Various flags;  see above for
				 * definitions. */
    Tcl_TimerToken timer;	/* Timer for grabbing one frame. */
    struct ng_devstate *dev;	/* LIBNG device. */
    struct ng_devstate devs[2];
    int wantFmt;		/* VIDEO_* format the widget needs. */
    struct ng_video_fmt fmt;	/* Current VIDEO_* format. */
    struct ng_process_handle *proc;	/* Format conversion or NULL. */
    struct ng_writer *writer;	/* Writer for recording. */
    void *writerHandle;		/* Handle of writer for recording. */
    struct ng_video_buf videoBuf;	/* Frame buffer. */
    int64_t lastT;
    double fps[2];		/* Requested/real frames per second. */
} V4LNG;

static Tk_ConfigSpec configSpecs[] = {
    {TK_CONFIG_BORDER, "-background", "background", "Background",
	"#d9d9d9", Tk_Offset(V4LNG, border), 0},
    {TK_CONFIG_SYNONYM, "-bd", "borderWidth", (char *) NULL,
	(char *) NULL, 0, 0},
    {TK_CONFIG_PIXELS, "-borderwidth", "borderWidth", "BorderWidth",
	"0", Tk_Offset(V4LNG, borderWidth), 0},
    {TK_CONFIG_SYNONYM, "-bg", "background", (char *) NULL,
       (char *) NULL, 0, 0},
    {TK_CONFIG_ACTIVE_CURSOR, "-cursor", "cursor", "Cursor",
	(char *) NULL, Tk_Offset(V4LNG, cursor), TK_CONFIG_NULL_OK},
    {TK_CONFIG_PIXELS, "-height", "height", "Height",
	"0", Tk_Offset(V4LNG, height), 0},
    {TK_CONFIG_COLOR, "-highlightbackground", "highlightBackground",
	"HighlightBackground", "#d9d9d9",
	Tk_Offset(V4LNG, highlightBgColorPtr), 0},
    {TK_CONFIG_COLOR, "-highlightcolor", "highlightColor", "HighlightColor",
	"#000000", Tk_Offset(V4LNG, highlightColorPtr), 0},
    {TK_CONFIG_PIXELS, "-highlightthickness", "highlightThickness",
	"HighlightThickness", "0", Tk_Offset(V4LNG, highlightWidth), 0},
#ifdef XV_SUPPORT
    {TK_CONFIG_BOOLEAN, "-noxv", "noXV",
	"NoXV", "0", Tk_Offset(V4LNG, noxv), 0},
#else
    {TK_CONFIG_BOOLEAN, "-noxv", "noXV",
	"NoXV", "1", Tk_Offset(V4LNG, noxv), 0},
#endif
    {TK_CONFIG_RELIEF, "-relief", "relief", "Relief",
	"flat", Tk_Offset(V4LNG, relief), 0},
    {TK_CONFIG_STRING, "-takefocus", "takeFocus", "TakeFocus",
	"0", Tk_Offset(V4LNG, takeFocus), TK_CONFIG_NULL_OK},
    {TK_CONFIG_PIXELS, "-width", "width", "Width",
	"0", Tk_Offset(V4LNG, width), 0},
    {TK_CONFIG_STRING, "-xscrollcommand", "xScrollCommand", "ScrollCommand",
	NULL, Tk_Offset(V4LNG, xScrollCmd), TK_CONFIG_NULL_OK},
    {TK_CONFIG_STRING, "-yscrollcommand", "yScrollCommand", "ScrollCommand",
	NULL, Tk_Offset(V4LNG, yScrollCmd), TK_CONFIG_NULL_OK},
    {TK_CONFIG_END, (char *) NULL, (char *) NULL, (char *) NULL,
	(char *) NULL, 0, 0}
};

/*
 * Forward declarations for procedures defined later in this file:
 */

static int	StopVideo _ANSI_ARGS_((V4LNG *v4lPtr, int stopTimer));
static int	StartVideo _ANSI_ARGS_((V4LNG *v4lPtr));
static int	CloseVideoDev _ANSI_ARGS_((V4LNG *v4lPtr));
static int	AttrWrite _ANSI_ARGS_((V4LNG *v4lPtr,
				       struct ng_attribute *attr, int val));
static void	SetFormat _ANSI_ARGS_((V4LNG *v4lPtr));
static struct ng_video_buf	*GetProcBuf _ANSI_ARGS_((void *handle,
							 struct ng_video_fmt *fmt));
static void	NextFrame _ANSI_ARGS_((ClientData clientData));
static int	HandleComplEvents _ANSI_ARGS_((ClientData clientData,
					       XEvent *eventPtr));
static int	ShmErrHandler _ANSI_ARGS_((ClientData clientData,
					   XErrorEvent *errEvent));
static int	V4lngCmd _ANSI_ARGS_((ClientData clientData,
				      Tcl_Interp *interp, int argc,
				      CONST char **argv));
static int	V4lngWidgetCmd _ANSI_ARGS_((ClientData clientData,
					    Tcl_Interp *interp,
					    int argc, CONST char **argv));
static void	DestroyV4lng _ANSI_ARGS_((char *memPtr));
static int	ConfigureV4lng _ANSI_ARGS_((Tcl_Interp *interp,
					    V4LNG *v4lPtr, int argc,
					    CONST char **argv, int flags));
static void	ComputeView _ANSI_ARGS_((V4LNG *v4lPtr));
static void	DisplayV4lng _ANSI_ARGS_((ClientData clientData));
static void	V4lngEventProc _ANSI_ARGS_((ClientData clientData,
					    XEvent *eventPtr));
static void	V4lngCmdDeletedProc _ANSI_ARGS_((ClientData clientData));

int		V4lng_Init _ANSI_ARGS_((Tcl_Interp *interp));


/*
 *--------------------------------------------------------------
 *
 * StopVideo --
 *
 *
 *--------------------------------------------------------------
 */

static int
StopVideo(v4lPtr, stopTimer)
    V4LNG *v4lPtr;
    int stopTimer;
{
    struct ng_devstate *dev = v4lPtr->dev;

    if (stopTimer) {
	Tcl_DeleteTimerHandler(v4lPtr->timer);
    }
    if ((dev != NULL) && (v4lPtr->flags & DEV_CAPTURE)) {
	dev->driver.v->stopvideo(dev->handle);
	v4lPtr->fps[1] = 0;
	v4lPtr->flags &= ~DEV_CAPTURE;
	return 1;
    }
    return 0;
}

/*
 *--------------------------------------------------------------
 *
 * StartVideo --
 *
 *
 *--------------------------------------------------------------
 */

static int
StartVideo(v4lPtr)
    V4LNG *v4lPtr;
{
    struct ng_devstate *dev = v4lPtr->dev;

    if ((dev != NULL) && !(v4lPtr->flags & DEV_CAPTURE)) {
	dev->driver.v->startvideo(dev->handle, (int) (1000 / v4lPtr->fps[0]),
				  4);
	v4lPtr->flags |= DEV_CAPTURE;
	Tcl_DeleteTimerHandler(v4lPtr->timer);
	v4lPtr->timer =
	    Tcl_CreateTimerHandler((int) (900 / v4lPtr->fps[0]),
				   NextFrame, (ClientData) v4lPtr);
	v4lPtr->flags |= (REDRAW_NOCLIP | UPDATE_XVIEW | UPDATE_YVIEW);
	v4lPtr->fps[1] = 0;
	v4lPtr->lastT = 0;
	if ((v4lPtr->width <= 0) || (v4lPtr->height <= 0)) {
	    int width, height;

	    width = v4lPtr->width;
	    if (width <= 0) {
		width = v4lPtr->fmt.width + 
		    2 * (v4lPtr->highlightWidth + v4lPtr->borderWidth);
	    }
	    height = v4lPtr->height;
	    if (height <= 0) {
		height = v4lPtr->fmt.height + 
		    2 * (v4lPtr->highlightWidth + v4lPtr->borderWidth);
	    }
	    Tk_GeometryRequest(v4lPtr->tkwin, width, height);
	    v4lPtr->flags |= UPDATE_XVIEW | UPDATE_YVIEW;
	}
	ComputeView(v4lPtr);
	return 1;
    }
    return 0;
}

/*
 *--------------------------------------------------------------
 *
 * CloseVideoDev --
 *
 *
 *--------------------------------------------------------------
 */

static int
CloseVideoDev(v4lPtr)
    V4LNG *v4lPtr;
{
    struct ng_devstate *dev = v4lPtr->dev;

    if (dev != NULL) {
	StopVideo(v4lPtr, 1);
	if (v4lPtr->proc != NULL) {
	    ng_process_fini(v4lPtr->proc);
	    v4lPtr->proc = NULL;
	}
	if (v4lPtr->flags & RECORDING) {
	    v4lPtr->writer->wr_close(v4lPtr->writerHandle);
	    v4lPtr->writerHandle = NULL;
	    v4lPtr->writer = NULL;
	}
	ng_dev_close(dev);
	ng_dev_fini(dev);
	v4lPtr->dev = NULL;
	v4lPtr->flags &= ~(DEV_CAPTURE | RECORDING);
	return 1;
    }
    return 0;
}

/*
 *--------------------------------------------------------------
 *
 * AttrWrite --
 *
 *
 *--------------------------------------------------------------
 */

static int
AttrWrite(v4lPtr, attr, val)
    V4LNG *v4lPtr;
    struct ng_attribute *attr;
    int val;
{
    int restart = 0;

    if (attr != NULL) {
	if ((attr->id == ATTR_ID_NORM) || (attr->id == ATTR_ID_INPUT)) {
	    restart = StopVideo(v4lPtr, 0);
	}
	attr->write(attr, val);
	if (restart) {
	    StartVideo(v4lPtr);
	}
	return attr->read(attr);
    }
    return -1;
}

/*
 *--------------------------------------------------------------
 *
 * SetFormat --
 *
 *
 *--------------------------------------------------------------
 */

static void
SetFormat(v4lPtr)
    V4LNG *v4lPtr;
{
    int i;
    struct ng_devstate *dev = v4lPtr->dev;
    struct ng_video_fmt fmt;
    struct ng_video_conv *conv;
#ifdef XV_SUPPORT
    int xv_id = 0;
#endif

    if (dev == NULL) {
	return;
    }
#ifdef XV_SUPPORT
    if (v4lPtr->hasxv) {
	if (v4lPtr->usexv && (v4lPtr->xvImg != NULL)) {
	    XFree(v4lPtr->xvImg);
	    v4lPtr->xvImg = NULL;
	}
	v4lPtr->usexv = 0;
	for (i = 0; i < v4lPtr->xvNfmts; i++) {
	    xv_id = v4lPtr->xvFmts[i].id;
	    if (xv_id == FOURCC('Y','U','Y','V')) {
		fmt.fmtid = VIDEO_YUYV;
	    } else if (xv_id == FOURCC('Y','U','Y','2')) {
		fmt.fmtid = VIDEO_YUYV;
	    } else if (xv_id == FOURCC('U','Y','V','Y')) {
		fmt.fmtid = VIDEO_UYVY;
	    } else if (xv_id == FOURCC('I','4','2','0')) {
		fmt.fmtid = VIDEO_YUV420P;
	    } else if (xv_id == FOURCC('Y','V','1','2')) {
		fmt.fmtid = VIDEO_YUV420P;
	    } else if ((v4lPtr->xvFmts[i].type == XvRGB) &&
		       (v4lPtr->xvFmts[i].format == XvPacked) &&
		       (v4lPtr->xvFmts[i].depth == Tk_Depth(v4lPtr->tkwin))) {
		fmt.fmtid = v4lPtr->wantFmt;
	    } else {
		continue;
	    }
	    fmt.width = v4lPtr->vwidth;
	    fmt.height = v4lPtr->vheight;
	    if (dev->driver.v->setformat(dev->handle, &fmt) >= 0) {
		v4lPtr->fmt = fmt;
		v4lPtr->usexv = 1;
		goto foundFmt;
	    }
	}
    }
#endif
    fmt.fmtid = v4lPtr->wantFmt;
    fmt.width = v4lPtr->vwidth;
    fmt.height = v4lPtr->vheight;
    if (dev->driver.v->setformat(dev->handle, &fmt) >= 0) {
	v4lPtr->fmt = fmt;
	goto foundFmt;
    }
    if (v4lPtr->wantFmt == VIDEO_RGB32) {
	fmt.fmtid = VIDEO_BGR32;
	fmt.width = v4lPtr->vwidth;
	fmt.height = v4lPtr->vheight;
	if (dev->driver.v->setformat(dev->handle, &fmt) >= 0) {
	    v4lPtr->fmt = fmt;
	    goto foundFmt;
	}
    }
    for (i = 0;; i++) {
	conv = ng_conv_find_to(v4lPtr->wantFmt, &i);
	if (conv == NULL) {
	    break;
	}
	fmt.fmtid = conv->fmtid_in;
	fmt.width = v4lPtr->vwidth;
	fmt.height = v4lPtr->vheight;
	if (dev->driver.v->setformat(dev->handle, &fmt) >= 0) {
	    v4lPtr->fmt = fmt;
	    v4lPtr->fmt.fmtid = v4lPtr->wantFmt;
	    if (v4lPtr->proc != NULL) {
		ng_process_fini(v4lPtr->proc);
		v4lPtr->proc = NULL;
	    }
	    v4lPtr->proc = ng_conv_init(conv, &fmt, &v4lPtr->fmt);
	    goto foundFmt;
	}
    }
    v4lPtr->fmt.fmtid = v4lPtr->wantFmt;
    v4lPtr->width = 1;
    v4lPtr->height = 1;
foundFmt:
    v4lPtr->fmt.bytesperline = v4lPtr->fmt.width * 4;
    v4lPtr->videoBuf.fmt = v4lPtr->fmt;
    v4lPtr->videoBuf.size = v4lPtr->fmt.width * v4lPtr->fmt.height * 4;
    v4lPtr->img->bytes_per_line = v4lPtr->fmt.width * 4;
    v4lPtr->img->width = v4lPtr->fmt.width;
    v4lPtr->img->height = v4lPtr->fmt.height;
    if (v4lPtr->proc != NULL) {
	ng_process_setup(v4lPtr->proc, GetProcBuf, (void *) v4lPtr);
    }
#ifdef XV_SUPPORT
    if (v4lPtr->usexv) {
	if (v4lPtr->islocal) {
	    v4lPtr->xvImg = XvShmCreateImage(v4lPtr->display, v4lPtr->xvPort,
					     xv_id, v4lPtr->img->data,
					     v4lPtr->img->width,
					     v4lPtr->img->height,
					     &v4lPtr->shmInfo);
	} else {
	    v4lPtr->xvImg = XvCreateImage(v4lPtr->display, v4lPtr->xvPort,
					  xv_id, v4lPtr->img->data,
					  v4lPtr->img->width,
					  v4lPtr->img->height);
	}
    }
#endif
}

/*
 *--------------------------------------------------------------
 *
 * GetProcBuf --
 *
 *
 *--------------------------------------------------------------
 */

static struct ng_video_buf *
GetProcBuf(handle, fmt)
    void *handle;
    struct ng_video_fmt *fmt;
{
    V4LNG *v4lPtr = (V4LNG *) handle;

    return &v4lPtr->videoBuf;
}

/*
 *--------------------------------------------------------------
 *
 * NextFrame --
 *
 *
 *--------------------------------------------------------------
 */

static void
NextFrame(clientData)
    ClientData clientData;
{
    V4LNG *v4lPtr = (V4LNG *) clientData;
    XRectangle rect;
    struct ng_devstate *dev = v4lPtr->dev;
    struct ng_video_buf *v, *v2 = NULL;
    int64_t diff;

    if (dev == NULL) {
	goto done;
    }
    v = dev->driver.v->nextframe(dev->handle);
    if (v == NULL) {
	goto done;
    }
    diff = v->info.ts - v4lPtr->lastT;
    if (diff > 0) {
	v4lPtr->fps[1] = (v4lPtr->fps[1] + 1E9 / diff) / 2;
    }
    v4lPtr->lastT = v->info.ts;
    if (v4lPtr->proc != NULL) {
	ng_process_put_frame(v4lPtr->proc, v);
	v2 = ng_process_get_frame(v4lPtr->proc);
    } else {
	v2 = &v4lPtr->videoBuf;
	v2->fmt = v->fmt;
	v2->size = v->size;
	v2->info = v->info;
	memcpy(v2->data, v->data, v->size);
    }
    if (v4lPtr->flags & RECORDING) {
	if (v4lPtr->writer->wr_video(v4lPtr->writerHandle, v2) < 0) {
	    v4lPtr->writer->wr_close(v4lPtr->writerHandle);
	    v4lPtr->writerHandle = NULL;
	    v4lPtr->writer = NULL;
	    v4lPtr->flags &= ~RECORDING;
	}
    }
    if (v4lPtr->proc == NULL) {
	ng_release_video_buf(v);
    }
    rect.x = v4lPtr->vxoff - v4lPtr->xoff;
    rect.y = v4lPtr->vyoff - v4lPtr->yoff;
#ifdef XV_SUPPORT
    if (v4lPtr->usexv) {
	rect.width = v4lPtr->viw;
	rect.height = v4lPtr->vih;
    } else {
	rect.width = v2->fmt.width;
	rect.height = v2->fmt.height;
    } 
#else    
    rect.width = v2->fmt.width;
    rect.height = v2->fmt.height;
#endif
    XUnionRectWithRegion(&rect, v4lPtr->region, v4lPtr->region);
    if (!(v4lPtr->flags & REDRAW_PENDING)) {
	Tcl_DoWhenIdle(DisplayV4lng, (ClientData) v4lPtr);
	v4lPtr->flags |= REDRAW_PENDING;
    }
done:
    if (v4lPtr->flags & DEV_CAPTURE) {
	v4lPtr->timer =
	    Tcl_CreateTimerHandler((int) (900 / v4lPtr->fps[0]),
				   NextFrame, (ClientData) v4lPtr);
    }
}

/*
 *--------------------------------------------------------------
 *
 * HandleComplEvents --
 *
 *
 *--------------------------------------------------------------
 */

static int
HandleComplEvents(clientData, eventPtr)
    ClientData clientData;
    XEvent *eventPtr;
{
    V4LNG *v4lPtr = (V4LNG *) clientData;
    Tk_Window tkwin = v4lPtr->tkwin;
    XShmCompletionEvent *shmEvent = (XShmCompletionEvent *) eventPtr;

#ifdef XV_SUPPORT
    if (v4lPtr->hasxv && (eventPtr->xany.type == v4lPtr->xvCompletion) &&
	(v4lPtr->shmInfo.shmseg == shmEvent->shmseg)) {
	goto doit;
    }
#endif
    if ((eventPtr->xany.type == v4lPtr->Completion) &&
	(v4lPtr->shmInfo.shmseg == shmEvent->shmseg)) {
#ifdef XV_SUPPORT
doit:
#endif
	if ((tkwin == NULL) || !Tk_IsMapped(tkwin)) {
	    v4lPtr->flags &= ~(COMPL_PENDING | REDRAW_AGAIN);
	    return 1;
	}
	v4lPtr->flags &= ~COMPL_PENDING;
	if (v4lPtr->flags & REDRAW_AGAIN) {
	    if (!(v4lPtr->flags & REDRAW_PENDING)) {
		Tcl_DoWhenIdle(DisplayV4lng, (ClientData) v4lPtr);
		v4lPtr->flags |= REDRAW_PENDING;
	    }
	    v4lPtr->flags &= ~REDRAW_AGAIN;
	}
	return 1;
    }
    return 0;
}

/*
 *--------------------------------------------------------------
 *
 * ShmErrHandler --
 *
 *	X error handler during XShm image creation.
 *
 *--------------------------------------------------------------
 */

static int
ShmErrHandler(clientData, errEvent)
    ClientData clientData;
    XErrorEvent *errEvent;
{
    int *goterr = (int *) clientData;

    *goterr = 1;
    return 0;
}

/*
 *--------------------------------------------------------------
 *
 * V4LNGCmd --
 *
 *	This procedure is invoked to process the "v4l"
 *	Tcl command.
 *
 * Results:
 *	A standard Tcl result.
 *
 * Side effects:
 *	See the user documentation.
 *
 *--------------------------------------------------------------
 */

static int
V4lngCmd(clientData, interp, argc, argv)
    ClientData clientData;	/* Main window associated with interpreter.
				 * If we're called by Tk_Init to create a
				 * new application, then this is NULL. */
    Tcl_Interp *interp;		/* Current interpreter. */
    int argc;			/* Number of arguments. */
    CONST char **argv;		/* Argument strings. */
{
    Tk_Window tkwin = (Tk_Window) clientData;
    V4LNG *v4lPtr;
    Tk_Window new;
    int islocal = 0;
#ifdef XV_SUPPORT
    int dummy;
#endif
    XGCValues gcValues;
    int mask, screen, fmt, endian;
    int paddrLen, laddrLen;
    struct sockaddr paddr, laddr;

    if (argc < 2) {
	Tcl_AppendResult(interp, "wrong # args: should be \"",
		argv[0], " pathName ?options?\"", (char *) NULL);
	return TCL_ERROR;
    }

    endian = 0xFF000000;
    endian = ((char *) &endian)[0];
    switch (Tk_Depth(tkwin)) {
    case 15:
	fmt = endian ? VIDEO_RGB15_BE : VIDEO_RGB15_LE;
	break;
    case 16:
	fmt = endian ? VIDEO_RGB16_BE : VIDEO_RGB16_LE;
	break;
    case 24:
    case 32:
	fmt = endian ? VIDEO_RGB32 : VIDEO_BGR32;
	break;
    default:
	Tcl_AppendResult(interp, "unsupported display depth", (char *) NULL);
	return TCL_ERROR;
    }

    /* Find out if we are on a local display. */
    paddrLen = sizeof (paddr);
    memset(&paddr, 0, sizeof (paddr));
    laddrLen = sizeof (laddr);
    memset(&laddr, 0, sizeof (laddr));
    if ((getpeername(ConnectionNumber(Tk_Display(tkwin)),
				      &paddr, &paddrLen) < 0) ||
	(getsockname(ConnectionNumber(Tk_Display(tkwin)),
				      &laddr, &laddrLen) < 0)) {
	Tcl_AppendResult(interp, "cannot determine X server connection: ",
			 Tcl_PosixError(interp), (char *) NULL);
	return TCL_ERROR;
    }
    switch (paddr.sa_family) {
    case AF_INET:
	if (((struct sockaddr_in *) &paddr)->sin_addr.s_addr ==
	    ((struct sockaddr_in *) &laddr)->sin_addr.s_addr) {
	    if (ntohs(((struct sockaddr_in *) &paddr)->sin_port) < 6010) {
		/* assume SSH starts at localhost:10 */
		islocal = 1;
	    }
	}
	break;
    case AF_UNIX:
	islocal = 1;
	break;
    }

    Tk_MakeWindowExist(tkwin);
    new = Tk_CreateWindowFromPath(interp, tkwin, argv[1], (char *) NULL);
    if (new == NULL) {
	goto error;
    }
    Tk_SetClass(new, "V4LNG");

    v4lPtr = (V4LNG *) Tcl_Alloc(sizeof (V4LNG));
    memset(v4lPtr, 0, sizeof (V4LNG));
    v4lPtr->islocal = islocal;
    v4lPtr->tkwin = new;
    v4lPtr->display = Tk_Display(new);
    v4lPtr->interp = interp;
    v4lPtr->widgetCmd = Tcl_CreateCommand(interp,
					  Tk_PathName(new),
					  (Tcl_CmdProc *)  V4lngWidgetCmd,
					  (ClientData) v4lPtr,
					  V4lngCmdDeletedProc);
    v4lPtr->cursor = None;
    v4lPtr->takeFocus = NULL;
    v4lPtr->border = NULL;
    v4lPtr->borderWidth = 0;
    v4lPtr->relief = TK_RELIEF_FLAT;
    v4lPtr->width = 0;
    v4lPtr->height = 0;
    v4lPtr->highlightWidth = 0;
    v4lPtr->highlightBgColorPtr = NULL;
    v4lPtr->highlightColorPtr = NULL;
    v4lPtr->iGC = None;
    v4lPtr->flags = 0;
    v4lPtr->xoff = v4lPtr->yoff = 0;
    v4lPtr->vwidth = 640;
    v4lPtr->vheight = 480;
    v4lPtr->vxoff = v4lPtr->vyoff = 0;
    v4lPtr->vih = v4lPtr->vih = 0;
    v4lPtr->wantFmt = fmt;
    v4lPtr->fps[0] = 25;
    v4lPtr->fps[1] = 0;

    screen = Tk_ScreenNumber(tkwin);
    v4lPtr->region = XCreateRegion();

    Tk_MakeWindowExist(new);

    gcValues.foreground = BlackPixel(v4lPtr->display,
				     Tk_ScreenNumber(v4lPtr->tkwin));
    gcValues.clip_x_origin = 0;
    gcValues.clip_y_origin = 0;
    mask = GCForeground | GCClipXOrigin | GCClipYOrigin;
    v4lPtr->iGC = XCreateGC(v4lPtr->display, Tk_WindowId(v4lPtr->tkwin),
			    mask, &gcValues);

#ifdef XV_SUPPORT
    v4lPtr->xvFmts = NULL;
    v4lPtr->xvImg = NULL;
    v4lPtr->xvPort = -1;
    v4lPtr->usexv = 0;
    v4lPtr->hasxv = XvQueryExtension(v4lPtr->display, &dummy,
				     &dummy, &dummy, &v4lPtr->xvCompletion,
				     &dummy) == Success;
#endif
    if (islocal && XShmQueryExtension(v4lPtr->display)) {
	Tk_ErrorHandler handler;
	int goterr = 0;

	handler = Tk_CreateErrorHandler(v4lPtr->display, -1, -1, -1,
					ShmErrHandler, (ClientData) &goterr);
	v4lPtr->Completion = XShmGetEventBase(v4lPtr->display) + ShmCompletion;
	if (goterr) {
	    Tk_DeleteErrorHandler(handler);
	    goto noshm_2;
	}
	v4lPtr->img = XShmCreateImage(v4lPtr->display,
				      DefaultVisual(v4lPtr->display, screen),
				      Tk_Depth(tkwin),
				      ZPixmap, NULL, &v4lPtr->shmInfo,
				      SHM_WIDTH, SHM_HEIGHT);
	XSync(v4lPtr->display, False);
	if (goterr) {
	    goto noshm_2;
	}
	v4lPtr->shmInfo.shmid = shmget(IPC_PRIVATE, SHM_SIZE,
				       IPC_CREAT | 0777);
	v4lPtr->shmInfo.shmaddr = (char *) shmat(v4lPtr->shmInfo.shmid, 0, 0);
	if (v4lPtr->shmInfo.shmaddr == (char *) -1) {
	    shmctl(v4lPtr->shmInfo.shmid, IPC_RMID, NULL);
	    goto noshm_1;
	}
	v4lPtr->shmInfo.readOnly = False;
	XShmAttach(v4lPtr->display, &v4lPtr->shmInfo);
	XSync(v4lPtr->display, False);
	shmctl(v4lPtr->shmInfo.shmid, IPC_RMID, NULL);
	if (goterr) {
	    shmdt(v4lPtr->shmInfo.shmaddr);
noshm_1:
	    XDestroyImage(v4lPtr->img);
noshm_2:	    
	    Tk_DeleteErrorHandler(handler);
	    goto noshm;
	}
	Tk_DeleteErrorHandler(handler);
	v4lPtr->img->data = v4lPtr->shmInfo.shmaddr;
    } else {
noshm:
	v4lPtr->islocal = 0;
	v4lPtr->img = XCreateImage(v4lPtr->display,
				   DefaultVisual(v4lPtr->display, screen),
				   Tk_Depth(tkwin),
				   ZPixmap, 0, NULL,
				   1, 1, 8, 0);
	v4lPtr->img->data = Tcl_Alloc(SHM_SIZE);
    }

    /* verify XImage endianess */
    switch (v4lPtr->img->depth) {
    case 15:
	if (v4lPtr->img->red_mask & 0x0007) {
	    fmt = endian ? VIDEO_RGB15_LE : VIDEO_RGB15_BE;
	}
	break;
    case 16:
	if (v4lPtr->img->red_mask & 0x0007) {
	    fmt = endian ? VIDEO_RGB16_BE : VIDEO_RGB16_LE;
	}
	break;
    case 24:
    case 32:
	if (v4lPtr->img->red_mask & 0x00FF) {
	    fmt = endian ? VIDEO_RGB32 : VIDEO_BGR32;
	}
	break;
    }
    v4lPtr->wantFmt = fmt;
    Tk_CreateEventHandler(new,
			  ExposureMask | StructureNotifyMask |
			  VisibilityChangeMask | FocusChangeMask,
			  V4lngEventProc, (ClientData) v4lPtr);
    if (ConfigureV4lng(interp, v4lPtr, argc - 2, argv + 2, 0) != TCL_OK) {
	goto error;
    }
    interp->result = Tk_PathName(new);
    if (v4lPtr->islocal) {
	Tk_CreateGenericHandler(HandleComplEvents, (ClientData) v4lPtr);
    }
    ng_init_video_buf(&v4lPtr->videoBuf);
    v4lPtr->videoBuf.data = v4lPtr->img->data;
#ifdef XV_SUPPORT
    if (v4lPtr->noxv) {
	v4lPtr->hasxv = 0;
    }
    if (v4lPtr->hasxv) {
	int i, k, nAdap = 0, nFmts, port = 0, found = 0;
	XvAdaptorInfo *adap = NULL;
	XvImageFormatValues *fmts = NULL;

	v4lPtr->xvCompletion += ShmCompletion;
	if (XvQueryAdaptors(v4lPtr->display, Tk_WindowId(new),
			    &nAdap, &adap) == Success) {
	    for (i = 0; i < nAdap; i++) {
		if (adap[i].type & XvImageMask) {
		    XvEncodingInfo *encs = NULL;

		    nFmts = adap[i].num_formats;
		    for (k = 0; k < nFmts; k++) {
			if (adap[i].formats &&
			    (adap[i].formats[k].visual_id ==
			     DefaultVisual(v4lPtr->display,
					   Tk_ScreenNumber(new))->visualid)) {
			    break;
			}
		    }
		    if (k >= nFmts) {
			goto skip;
		    }
		    if (adap[i].num_ports < 1) {
			goto skip;
		    }
		    port = adap[i].base_id;
		    if ((XvQueryEncodings(v4lPtr->display, port,
					  &nFmts, &encs) != Success) ||
			(nFmts == 0)) {
			goto skip;
		    }
		    for (k = 0; k < nFmts; k++) {
			if (strcmp(encs[k].name, "XV_IMAGE") == 0) {
			    v4lPtr->xvMaxWidth = encs[k].width;
			    v4lPtr->xvMaxHeight = encs[k].height;
			    break;
			}
		    }
		    XvFreeEncodingInfo(encs);
		    if (k >= nFmts) {
			goto skip;
		    }
		    for (k = 0; k < adap[i].num_ports; k++) {
			port = adap[i].base_id + k;
			if (XvGrabPort(v4lPtr->display, port,
				       CurrentTime) == Success) {
			    v4lPtr->xvPort = port;
			    break;
			}
		    }
		    if (k >= adap[i].num_ports) {
			goto skip;
		    }
		    fmts = XvListImageFormats(v4lPtr->display,
					      adap[i].base_id, &nFmts);
		    if (fmts && (nFmts > 0)) {
			found = 1;
			break;
		    }
		    if (fmts != NULL) {
			XFree(fmts);
			fmts = NULL;
		    }
		}
	    }
skip:
	    XvFreeAdaptorInfo(adap);
	}
	if (found) {
	    v4lPtr->xvNfmts = nFmts;
	    v4lPtr->xvFmts = fmts;
	} else {
	    v4lPtr->hasxv = 0;
	    if (v4lPtr->xvPort != -1) {
		XvUngrabPort(v4lPtr->display, v4lPtr->xvPort, CurrentTime);
	    }
	}
    }
#endif
    return TCL_OK;

error:
    if (new != NULL) {
	Tk_DestroyWindow(new);
    }
    return TCL_ERROR;
}

/*
 *--------------------------------------------------------------
 *
 * V4LNGWidgetCmd --
 *
 *	This procedure is invoked to process the Tcl command
 *	that corresponds to a v4l widget.  See the user
 *	documentation for details on what it does.
 *
 * Results:
 *	A standard Tcl result.
 *
 * Side effects:
 *	See the user documentation.
 *
 *--------------------------------------------------------------
 */

static int
V4lngWidgetCmd(clientData, interp, argc, argv)
    ClientData clientData;	/* Information about v4l widget. */
    Tcl_Interp *interp;		/* Current interpreter. */
    int argc;			/* Number of arguments. */
    CONST char **argv;		/* Argument strings. */
{
    V4LNG *v4lPtr = (V4LNG *) clientData;
    int result = TCL_OK;
    size_t length;
    int c;
    static char *drivers[] = { "v4l", "v4l2", "bktr" };

    if (argc < 2) {
	Tcl_AppendResult(interp, "wrong # args: should be \"",
			 argv[0], " option ?arg arg ...?\"", (char *) NULL);
	return TCL_ERROR;
    }
    Tcl_Preserve((ClientData) v4lPtr);
    c = argv[1][0];
    length = strlen(argv[1]);
    if ((c == 'c') && (strncmp(argv[1], "cget", length) == 0) &&
	(length >= 2)) {
	if (argc != 3) {
	    Tcl_AppendResult(interp, "wrong # args: should be \"",
			     argv[0], " cget option\"", (char *) NULL);
	    result = TCL_ERROR;
	    goto done;
	}
	result = Tk_ConfigureValue(interp, v4lPtr->tkwin, configSpecs,
				   (char *) v4lPtr, argv[2], 0);
    } else if ((c == 'c') && (strncmp(argv[1], "configure", length) == 0) &&
	       (length >= 4)) {
	if (argc == 2) {
	    result = Tk_ConfigureInfo(interp, v4lPtr->tkwin, configSpecs,
				      (char *) v4lPtr, (char *) NULL, 0);
	} else if (argc == 3) {
	    result = Tk_ConfigureInfo(interp, v4lPtr->tkwin, configSpecs,
				      (char *) v4lPtr, argv[2], 0);
	} else {
	    result = ConfigureV4lng(interp, v4lPtr, argc - 2, argv + 2,
				  TK_CONFIG_ARGV_ONLY);

	}
    } else if ((c == 'c') && (strncmp(argv[1], "capture", length) == 0) &&
	       (length >= 3)) {
	if (argc == 2) {
	    strcpy(interp->result, (v4lPtr->flags & DEV_CAPTURE) ? "1" : "0");
	    goto done;
	}
	if (argc != 3) {
	    Tcl_AppendResult(interp, "wrong # args: should be \"",
			     argv[0], " capture ?bool?\"", (char *) NULL);
	    result = TCL_ERROR;
	    goto done;
	}
	if (Tcl_GetBoolean(interp, argv[2], &c) != TCL_OK) {
	    result = TCL_ERROR;
	    goto done;
	}
	if (v4lPtr->dev == NULL) {
	    goto done;
	}
	if (c) {
	    StartVideo(v4lPtr);
	} else {
	    StopVideo(v4lPtr, 1);
	    if (!(v4lPtr->flags & REDRAW_PENDING)) {
		Tcl_DoWhenIdle(DisplayV4lng, (ClientData) v4lPtr);
		v4lPtr->flags |= REDRAW_PENDING;
	    }
	}
    } else if ((c == 'f') && (strncmp(argv[1], "fps", length) == 0) &&
	       (length >= 2)) {
	int rate;
	char buf[TCL_DOUBLE_SPACE * 3];

	if (argc > 3) {
	    Tcl_AppendResult(interp, "wrong # args: should be \"",
			     argv[0], " fps ?rate?", (char *) NULL);
	    result = TCL_ERROR;
	    goto done;
	}
	if (argc == 2) {
	    Tcl_PrintDouble(interp, v4lPtr->fps[0], buf);
	    strcat(buf, " ");
	    Tcl_PrintDouble(interp, v4lPtr->fps[1], buf + strlen(buf));
	    Tcl_AppendResult(interp, buf, (char *) NULL);
	} else {
	    if (Tcl_GetInt(interp, argv[2], &rate) != TCL_OK) {
		result = TCL_ERROR;
		goto done;
	    }
	    if (rate < 1) {
		rate = 1;
	    } else if (rate > 50) {
		rate = 50;
	    }
	    if (!(v4lPtr->flags & RECORDING)) {
		v4lPtr->fps[0] = rate;
	    }
	}
    } else if ((c == 'g') && (strncmp(argv[1], "geometry", length) == 0) &&
	       (length >= 3)) {
	int neww, newh, restart = 0;

	if ((argc != 2) && (argc != 4)) {
	    Tcl_AppendResult(interp, "wrong # args: should be \"",
			     argv[0], " ?width height?", (char *) NULL);
	    result = TCL_ERROR;
	    goto done;
	}
	if (argc == 2) {
	    sprintf(interp->result, "%d %d", v4lPtr->vwidth, v4lPtr->vheight);
	} else {
	    if (Tcl_GetInt(interp, argv[2], &neww) != TCL_OK) {
		result = TCL_ERROR;
		goto done;
	    }
	    if (Tcl_GetInt(interp, argv[3], &newh) != TCL_OK) {
		result = TCL_ERROR;
		goto done;
	    }
	    if ((neww <= 0 || neww > SHM_WIDTH) ||
		(newh <= 0 || newh > SHM_HEIGHT) ||
		(neww * newh <= 0)) {
		Tcl_AppendResult(interp, "illegal geometry", (char *) NULL);
		result = TCL_ERROR;
		goto done;
	    }
	    if (v4lPtr->flags & RECORDING) {
		/* ignore while recording */
		goto done;
	    }
	    v4lPtr->vwidth = neww;
	    v4lPtr->vheight = newh;
	    if (v4lPtr->flags & DEV_CAPTURE) {
		restart = StopVideo(v4lPtr, 0);
	    }
	    if (v4lPtr->dev != NULL) {
		SetFormat(v4lPtr);
	    }
	    if (restart) {
		StartVideo(v4lPtr);
	    }
	}
    } else if ((c == 'g') && (strncmp(argv[1], "getattr", length) == 0) &&
	       (length >= 3)) {
	int i;
	struct ng_attribute *attr;
		       
	if (argc > 3) {
	    Tcl_AppendResult(interp, "wrong # args: should be \"",
			     argv[0], " getattr ?name?",
			     (char *) NULL);
	    result = TCL_ERROR;
	    goto done;
	}
	if (v4lPtr->dev != NULL) {
	    if (argc == 2) {
		int *ids = NULL, num;

		num = ng_attr_enum(v4lPtr->dev, NULL, 0);
		if (num > 0) {
		    ids = (int *) Tcl_Alloc(num * sizeof (int));
		    num = ng_attr_enum(v4lPtr->dev, ids, num);
		}
		for (i = 0; (ids != NULL) && (i < num); i++) {
		    attr = ng_attr_byid(v4lPtr->dev, ids[i]);
		    if (attr) {
			Tcl_AppendElement(interp, attr->name);
		    }
		}
		if (ids != NULL) {
		    Tcl_Free((char *) ids);
		}
	    } else {
		attr = ng_attr_byname(v4lPtr->dev, (char *) argv[2]);
		if (attr != NULL) {
		    int val = attr->read(attr);
		    char vbuf[64];

		    switch (attr->type) {
		    case ATTR_TYPE_INTEGER:
			sprintf(vbuf, "%d", val);
			Tcl_AppendElement(interp, vbuf);
			Tcl_AppendElement(interp, "integer");
			sprintf(vbuf, "%d", attr->min);
			Tcl_AppendElement(interp, vbuf);
			sprintf(vbuf, "%d", attr->max);
			Tcl_AppendElement(interp, vbuf);
			sprintf(vbuf, "%d", attr->defval);
			Tcl_AppendElement(interp, vbuf);
			break;
		    case ATTR_TYPE_BOOL:
			strcpy(vbuf, val ? "1" : "0");
			Tcl_AppendElement(interp, vbuf);
			Tcl_AppendElement(interp, "boolean");
			Tcl_AppendElement(interp, "0");
			Tcl_AppendElement(interp, "1");
			strcpy(vbuf, attr->defval ? "1" : "0");
			Tcl_AppendElement(interp, vbuf);
			break;
		    case ATTR_TYPE_CHOICE:
			if (attr->choices == NULL) {
			    break;
			}
			i = 0;
			while (attr->choices[i].str != NULL) {
			    if (attr->choices[i].nr == val) {
				Tcl_AppendElement(interp,
						  attr->choices[i].str);

				break;
			    }
			    ++i;
			}
			if (attr->choices[i].str == NULL) {
			    /* not found, report default */
			    i = 0;
			    while (attr->choices[i].str != NULL) {
				if (attr->choices[i].nr == attr->defval) {
				    Tcl_AppendElement(interp,
						      attr->choices[i].str);

				    break;
				}
				++i;
			    }
			    if (attr->choices[i].str == NULL) {
				/* not found, report empty choice */
				Tcl_AppendElement(interp, "");
			    }
			}
			Tcl_AppendElement(interp, "choice");
			i = 0;
			while (attr->choices[i].str != NULL) {
			    Tcl_AppendElement(interp,
					      attr->choices[i].str);
			    ++i;
			}
			break;
		    }
		}
	    }
	}
    } else if ((c == 's') && (strncmp(argv[1], "setattr", length) == 0) &&
	       (length >= 2)) {
	int i;
	struct ng_attribute *attr;
		       
	if ((argc < 3) || (argc > 4)) {
	    Tcl_AppendResult(interp, "wrong # args: should be \"",
			     argv[0], " setattr name ?value?",
			     (char *) NULL);
	    result = TCL_ERROR;
	    goto done;
	}
	if (v4lPtr->dev != NULL) {
	    attr = ng_attr_byname(v4lPtr->dev, (char *) argv[2]);
	    if (attr != NULL) {
		int val = attr->read(attr), newval;
		char vbuf[64];

		switch (attr->type) {
		case ATTR_TYPE_INTEGER:
		    if (argc == 3) {
			sprintf(vbuf, "%d", val);
			Tcl_AppendElement(interp, vbuf);
			Tcl_AppendElement(interp, "integer");
			sprintf(vbuf, "%d", attr->min);
			Tcl_AppendElement(interp, vbuf);
			sprintf(vbuf, "%d", attr->max);
			Tcl_AppendElement(interp, vbuf);
		    } else {
			if (Tcl_GetInt(interp, argv[3], &newval) != TCL_OK) {
			    result = TCL_ERROR;
			    goto done;
			}
			val = AttrWrite(v4lPtr, attr, newval);
			sprintf(vbuf, "%d", val);
			Tcl_AppendElement(interp, vbuf);
		    }
		    break;
		case ATTR_TYPE_BOOL:
		    if (argc == 3) {
			strcpy(vbuf, val ? "1" : "0");
			Tcl_AppendElement(interp, vbuf);
			Tcl_AppendElement(interp, "boolean");
			Tcl_AppendElement(interp, "0");
			Tcl_AppendElement(interp, "1");
		    } else {
			if (Tcl_GetBoolean(interp, argv[3], &newval)
			    != TCL_OK) {
			    result = TCL_ERROR;
			    goto done;
			}
			val = AttrWrite(v4lPtr, attr, newval);
			strcpy(vbuf, val ? "1" : "0");
			Tcl_AppendElement(interp, vbuf);
		    }
		    break;
		case ATTR_TYPE_CHOICE:
		    if (argc == 3) {
			i = 0;
			while (attr->choices[i].str != NULL) {
			    if (attr->choices[i].nr == val) {
				Tcl_AppendElement(interp,
						  attr->choices[i].str);

				break;
			    }
			    ++i;
			}
			if (attr->choices[i].str == NULL) {
			    Tcl_AppendElement(interp, "");
			}
			Tcl_AppendElement(interp, "choice");
			i = 0;
			while (attr->choices[i].str != NULL) {
			    Tcl_AppendElement(interp,
					      attr->choices[i].str);
			    ++i;
			}
		    } else {
			length = strlen(argv[3]);
			if (length < 1) {
			    length = 1;
			}
			i = 0;
			while (attr->choices[i].str != NULL) {
			    if (strncmp(attr->choices[i].str, argv[3],
					length) == 0) {
				newval = attr->choices[i].nr;
				break;
			    }
			    ++i;
			}
			if (attr->choices[i].str != NULL) {
			    val = AttrWrite(v4lPtr, attr, newval);
			} else {
			    val = attr->read(attr);
			}
			i = 0;
			while (attr->choices[i].str != NULL) {
			    if (attr->choices[i].nr == val) {
				Tcl_AppendElement(interp,
						  attr->choices[i].str);

				break;
			    }
			    ++i;
			}
			if (attr->choices[i].str == NULL) {
			    Tcl_AppendElement(interp, "");
			}
		    }
		    break;
		}
	    }
	}
    } else if ((c == 'd') && (strncmp(argv[1], "devices", length) == 0) &&
	       (length >= 3)) {
	int i = -1, k;

	if (argc != 2) {
	    Tcl_AppendResult(interp, "wrong # args: should be \"",
			     argv[0], " devices", (char *) NULL);
	    result = TCL_ERROR;
	    goto done;
	}
	for (k = 0; k < sizeof (drivers) / sizeof (drivers[0]); k++) {
	    struct ng_devinfo *info;

	    i = -1;
	    info = ng_vid_probe(drivers[k]);
	    if (info == NULL) {
		continue;
	    }
	    for (i = 0; info[i].device[0] != '\0'; i++) {
		Tcl_AppendElement(interp, info[i].device);
		Tcl_AppendElement(interp, info[i].name);
	    }
	    free(info);
	}
    } else if ((c == 'o') && (strncmp(argv[1], "open", length) == 0) &&
	       (length >= 2)) {
	int i = -1, k;
	char devname[64];
	struct ng_devstate *newdev = &v4lPtr->devs[0];

	if (v4lPtr->dev == &v4lPtr->devs[0]) {
	    newdev = &v4lPtr->devs[1];
	}
	if (argc != 3) {
	    Tcl_AppendResult(interp, "wrong # args: should be \"",
			     argv[0], " open device", (char *) NULL);
	    result = TCL_ERROR;
	    goto done;
	}
	devname[0] = '\0';
	for (k = 0; k < sizeof (drivers) / sizeof (drivers[0]); k++) {
	    struct ng_devinfo *info;

	    i = -1;
	    info = ng_vid_probe(drivers[k]);
	    if (info == NULL) {
		continue;
	    }
	    for (i = 0; info[i].device[0] != '\0'; i++) {
		if (strcmp(argv[2], info[i].device) == 0) {
		    strncpy(devname, info[i].device, sizeof (devname));
		    free(info);
		    goto devFound;
		}
		if (strcmp(argv[2], info[i].name) == 0) {
		    strncpy(devname, info[i].device, sizeof (devname));
		    free(info);
		    goto devFound;
		}
	    }
	    free(info);
	}
devFound:
	devname[sizeof (devname) - 1] = '\0';
	if (!devname[0] || ng_vid_init(newdev, devname)) {
	    Tcl_AppendResult(interp, "device \"", argv[2], "\" unavailable",
			     (char *) NULL);
	    result = TCL_ERROR;
	    goto done;
	}
	if (!(newdev->flags & CAN_CAPTURE)) {
	    ng_dev_fini(newdev);
	    Tcl_AppendResult(interp, "device \"", devname, "\" can't capture",
			     (char *) NULL);
	    result = TCL_ERROR;
	    goto done;
	}
	CloseVideoDev(v4lPtr);
	ng_dev_open(newdev);
	v4lPtr->dev = newdev;
	SetFormat(v4lPtr);
    } else if ((c == 'd') && (strncmp(argv[1], "debug", length) == 0) &&
	       (length >= 3)) {
	int level;

	if (argc == 2) {
	    sprintf(interp->result, "%d", ng_debug);
	    goto done;
	}
	if (argc != 3) {
	    Tcl_AppendResult(interp, "wrong # args: should be \"",
			     argv[0], " debug ?level?", (char *) NULL);
	    result = TCL_ERROR;
	    goto done;
	}
	if (Tcl_GetInt(interp, argv[2], &level) != TCL_OK) {
	    result = TCL_ERROR;
	    goto done;
	}
	ng_debug = level;
    } else if ((c == 'c') && (strncmp(argv[1], "close", length) == 0) &&
	       (length >= 2)) {
	if (argc != 2) {
	    Tcl_AppendResult(interp, "wrong # args: should be \"",
			     argv[0], " close", (char *) NULL);
	    result = TCL_ERROR;
	    goto done;
	}
	CloseVideoDev(v4lPtr);
	if (!(v4lPtr->flags & REDRAW_PENDING)) {
	    Tcl_DoWhenIdle(DisplayV4lng, (ClientData) v4lPtr);
	    v4lPtr->flags |= REDRAW_PENDING;
	}
    } else if ((c == 'r') && (strncmp(argv[1], "record", length) == 0) &&
	       (length >= 2)) {
	struct ng_writer *writer;
	void *whandle;
	struct ng_audio_fmt nulaudio = { AUDIO_NONE, 0 };
	int i;

	if (argc < 3) {
	    strcpy(interp->result, (v4lPtr->flags & RECORDING) ? "1" : "0");
	    goto done;
	}
	if ((argc == 3) && argv[2][0] && (strncmp(argv[2], "stop", 4) == 0)) {
	    if (v4lPtr->flags & RECORDING) {
		v4lPtr->writer->wr_close(v4lPtr->writerHandle);
		v4lPtr->writerHandle = NULL;
		v4lPtr->writer = NULL;
		v4lPtr->flags &= ~RECORDING;
	    }
	    goto done;
	}
	if (argc != 4) {
	    Tcl_AppendResult(interp, "wrong # args: should be \"",
			     argv[0], " record ?stop | fmt filename?",
			     (char *) NULL);
	    result = TCL_ERROR;
	    goto done;
	}
	writer = ng_find_writer_name((char *) argv[2]);
	if (writer == NULL) {
	    Tcl_AppendResult(interp, "format \"", argv[2], "\" not found",
			     (char *) NULL);
	    result = TCL_ERROR;
	    goto done;
	}
	for (i = 0; writer->video[i].name != NULL; i++) {
	    if (writer->video[i].fmtid == v4lPtr->fmt.fmtid) {
		break;
	    }
	}
	if (writer->video[i].name == NULL) {
	    Tcl_AppendResult(interp, "no compatible video format",
			     (char *) NULL);
	    result = TCL_ERROR;
	    goto done;
	}
	whandle = writer->wr_open((char *) argv[3], NULL, &v4lPtr->fmt,
				  (const void *) writer->video[i].priv,
				  (int) v4lPtr->fps[0] * 1000,
				  &nulaudio, (const void *) 0);
	if (whandle != NULL) {
	    if (v4lPtr->flags & RECORDING) {
		v4lPtr->writer->wr_close(v4lPtr->writerHandle);
		v4lPtr->writerHandle = NULL;
		v4lPtr->writer = NULL;
		v4lPtr->flags &= ~RECORDING;
	    }
	    v4lPtr->writer = writer;
	    v4lPtr->writerHandle = whandle;
	    v4lPtr->flags |= RECORDING;
	} else {
	    Tcl_AppendResult(interp, "unable to open writer/file",
			     (char *) NULL);
	    result = TCL_ERROR;
	    goto done;
	}
    } else if ((c == 'p') && (strncmp(argv[1], "photo", length) == 0) &&
	       (length >= 1)) {
	Tk_PhotoImageBlock block;
	Tk_PhotoHandle ph;
	int restart;
	struct ng_devstate *dev = v4lPtr->dev;
	struct ng_video_buf *v, *v2 = NULL;
	struct ng_process_handle *proc = NULL;

	if (argc != 3) {
	    Tcl_AppendResult(interp, "wrong # args: should be \"",
			     argv[0], " photo name\"", (char *) NULL);
	    result = TCL_ERROR;
	    goto done;
	}
	ph = Tk_FindPhoto(interp, argv[2]);
	if (ph == NULL) {
	    result = TCL_ERROR;
	    goto done;
	}
	if (dev == NULL) {
	    Tcl_AppendResult(interp, "no device opened", (char *) NULL);
	    result = TCL_ERROR;
	    goto done;
	}
	if (v4lPtr->fmt.fmtid == VIDEO_RGB32) {
	    block.offset[0] = 0;
	    block.offset[1] = 1;
	    block.offset[2] = 2;
	    block.offset[3] = -1;
	} else if (v4lPtr->fmt.fmtid == VIDEO_BGR32) {
	    block.offset[0] = 2;
	    block.offset[1] = 1;
	    block.offset[2] = 0;
	    block.offset[3] = -1;
	} else if (v4lPtr->proc == NULL) {
	    struct ng_video_fmt pfmt;
	    struct ng_video_conv *conv;

	    pfmt = v4lPtr->fmt;
	    pfmt.fmtid = VIDEO_RGB32;
	    conv = ng_conv_find_match(v4lPtr->fmt.fmtid, pfmt.fmtid);
	    if (conv != NULL) {
		block.offset[0] = 0;
		block.offset[1] = 1;
		block.offset[2] = 2;
		block.offset[3] = -1;
	    } else {
		pfmt.fmtid = VIDEO_BGR32;
		conv = ng_conv_find_match(v4lPtr->fmt.fmtid, pfmt.fmtid);
		if (conv != NULL) {
		    block.offset[0] = 2;
		    block.offset[1] = 1;
		    block.offset[2] = 0;
		    block.offset[3] = -1;
		}
	    }
	    if (conv == NULL) {
		goto incompat;
	    }
	    proc = ng_conv_init(conv, &v4lPtr->fmt, &pfmt);
	    ng_process_setup(proc, GetProcBuf, (void *) v4lPtr);
	} else {
incompat:
	    Tcl_AppendResult(interp, "incompatible video format",
			     (char *) NULL);
	    result = TCL_ERROR;
	    goto done;
	}
	restart = StopVideo(v4lPtr, 1);
	dev->driver.v->startvideo(dev->handle, 1, 4);
	v = dev->driver.v->nextframe(dev->handle);
	if (v != NULL) {
	    if (proc != NULL) {
		ng_process_put_frame(proc, v);
		v2 = ng_process_get_frame(proc);
	    } else if (v4lPtr->proc != NULL) {
		ng_process_put_frame(v4lPtr->proc, v);
		v2 = ng_process_get_frame(v4lPtr->proc);
	    } else {
		v2 = &v4lPtr->videoBuf;
		v2->fmt = v->fmt;
		v2->size = v->size;
		v2->info = v->info;
		memcpy(v2->data, v->data, v->size);
	    }
	    if ((proc == NULL) && (v4lPtr->proc == NULL)) {
		ng_release_video_buf(v);
	    }
	    block.pixelPtr = v4lPtr->img->data;
	    block.width = v4lPtr->img->width;
	    block.height = v4lPtr->img->height;
	    block.pitch = v4lPtr->img->bytes_per_line;
	    block.pixelSize = 4;
	    Tk_PhotoExpand(ph, block.width, block.height);
	    Tk_PhotoPutBlock(ph, &block, 0, 0, block.width, block.height,
			     TK_PHOTO_COMPOSITE_SET);
	} else {
	    Tcl_AppendResult(interp, "video capture failed", (char *) NULL);
	    result = TCL_ERROR;
	}
	dev->driver.v->stopvideo(dev->handle);
	if (proc != NULL) {
	    ng_process_fini(proc);
	}
	if (restart) {
	    StartVideo(v4lPtr);
	}
    } else if ((c == 's') && (strncmp(argv[1], "scan", length) == 0) &&
	       (length >= 2)) {
	int x, y, dx, dy, gain = 1;

	if ((argc != 5) && (argc != 6)) {
	    Tcl_AppendResult(interp, "wrong # args: should be \"",
			     argv[0], " scan mark x y\" or \"",
			     argv[0], " scan dragto x y ?gain?\"",
			     (char *) NULL);
	    result = TCL_ERROR;
	    goto done;
	}
	if ((Tcl_GetInt(interp, argv[3], &x) != TCL_OK) ||
	    (Tcl_GetInt(interp, argv[4], &y) != TCL_OK)) {
	    result = TCL_ERROR;
	    goto done;
	}
	if ((argc == 6) && (Tcl_GetInt(interp, argv[5], &gain) != TCL_OK)) {
	    result = TCL_ERROR;
	    goto done;
	}
	c = argv[2][0];
	length = strlen(argv[2]);
	if ((c == 'd') && (strncmp(argv[2], "dragto", length) == 0)) {
	    /* code after if-else block */
	} else if ((c == 'm') && (strncmp(argv[2], "mark", length) == 0)) {
	    v4lPtr->scanMarkX = x;
	    v4lPtr->scanMarkY = y;
	    goto done;
	} else {
	    Tcl_AppendResult(interp, "bad scan option \"", argv[2],
			     "\": must be mark or dragto", (char *) NULL);
	    result = TCL_ERROR;
	    goto done;
	}
	dx = gain * (v4lPtr->scanMarkX - x);
	dy = gain * (v4lPtr->scanMarkY - y);
	v4lPtr->scanMarkX = x;
	v4lPtr->scanMarkY = y;
	if ((dx != 0) || (dy != 0)) {
	    if (v4lPtr->viw < v4lPtr->img->width) {
		int xoff;

		xoff = v4lPtr->xoff + dx;
		if (xoff < 0) {
		    xoff = 0;
		} else if (xoff + v4lPtr->viw > v4lPtr->img->width) {
		    xoff = v4lPtr->img->width - v4lPtr->viw;
		}
		if (v4lPtr->xoff != xoff) {
		    v4lPtr->xoff = xoff;
		    v4lPtr->flags |= UPDATE_XVIEW;
		}
	    }
	    if (v4lPtr->vih < v4lPtr->img->height) {
		int yoff;

		yoff = v4lPtr->yoff + dy;
		if (yoff < 0) {
		    yoff = 0;
		} else if (yoff + v4lPtr->vih > v4lPtr->img->height) {
		    yoff = v4lPtr->img->height - v4lPtr->vih;
		}
		if (v4lPtr->yoff != yoff) {
		    v4lPtr->yoff = yoff;
		    v4lPtr->flags |= UPDATE_YVIEW;
		}
	    }
	    if (v4lPtr->flags & (UPDATE_XVIEW | UPDATE_YVIEW)) {
		v4lPtr->flags |= REDRAW_NOCLIP;
		ComputeView(v4lPtr);
		goto redraw;
	    }
	}
    } else if ((c == 'x') && (strncmp(argv[1], "xview", length) == 0) &&
	       (length >= 1)) {
	if (argc == 2) {
	    char buffer[128];
	    double amt[2];

	    if ((v4lPtr->viw <= 0) || (v4lPtr->viw >= v4lPtr->img->width)) {
		amt[0] = 0;
		amt[1] = 1;
	    } else {
		amt[0] = amt[1] = (double) v4lPtr->xoff /
		    (double) v4lPtr->img->width;
		amt[1] += (double) v4lPtr->viw / (double) v4lPtr->img->width;
	    }
	    Tcl_PrintDouble(interp, amt[0], buffer);
	    Tcl_AppendElement(interp, buffer);
	    Tcl_PrintDouble(interp, amt[1], buffer);
	    Tcl_AppendElement(interp, buffer);
	} else if (argc == 3) {
	    int xoff;

	    if (Tcl_GetInt(interp, argv[2], &xoff) == TCL_OK) {
		if (v4lPtr->viw < v4lPtr->img->width) {
		    if (xoff < 0) {
			xoff = 0;
		    } else if (xoff + v4lPtr->viw > v4lPtr->img->width) {
			xoff = v4lPtr->img->width - v4lPtr->viw;
		    }
		    v4lPtr->xoff = xoff;
		    ComputeView(v4lPtr);
		    v4lPtr->flags |= REDRAW_NOCLIP | UPDATE_XVIEW;
		    goto redraw;
		}
	    }
	} else if (argc == 4) {
	    double frac;
	    int xoff;

	    if (strcmp(argv[2], "moveto")) {
		goto badxview;
	    }
	    if (Tcl_GetDouble(interp, argv[3], &frac) == TCL_OK) {
		if (v4lPtr->viw < v4lPtr->img->width) {
		    xoff = frac * v4lPtr->img->width;
		    if (xoff < 0) {
			xoff = 0;
		    } else if (xoff + v4lPtr->viw > v4lPtr->img->width) {
			xoff = v4lPtr->img->width - v4lPtr->viw;
		    }
		    v4lPtr->xoff = xoff;
		    ComputeView(v4lPtr);
		    v4lPtr->flags |= REDRAW_NOCLIP | UPDATE_XVIEW;
		    goto redraw;
		}
	    }
	} else if (argc == 5) {
	    int n, isunit = 0;

	    c = argv[4][0];
	    length = strlen(argv[4]);
	    if (strcmp(argv[2], "scroll")) {
		goto badxview;
	    }
	    if ((c == 'u') && (strncmp(argv[4], "units", length) == 0)) {
		isunit = 1;
	    } else if ((c != 'p') || strncmp(argv[4], "pages", length)) {
		goto badxview;
	    }
	    if (Tcl_GetInt(interp, argv[3], &n) == TCL_OK) {
		if (v4lPtr->viw < v4lPtr->img->width) {
		    int xoff;

		    if (isunit) {
			xoff = v4lPtr->xoff + n;
		    } else {
			xoff = v4lPtr->xoff + n * v4lPtr->img->width / 2;
		    }
		    if (xoff < 0) {
			xoff = 0;
		    } else if (xoff + v4lPtr->viw > v4lPtr->img->width) {
			xoff = v4lPtr->img->width - v4lPtr->viw;
		    }
		    v4lPtr->xoff = xoff;
		    ComputeView(v4lPtr);
		    v4lPtr->flags |= REDRAW_NOCLIP | UPDATE_XVIEW;
		    goto redraw;
		}
	    }
	} else {
badxview:
	    Tcl_AppendResult(interp, "bad args to xview subcommand",
			     (char *) NULL);
	    result = TCL_ERROR;
	}
    } else if ((c == 'y') && (strncmp(argv[1], "yview", length) == 0) &&
	       (length >= 1)) {
	if (argc == 2) {
	    char buffer[128];
	    double amt[2];

	    if ((v4lPtr->vih <= 0) || (v4lPtr->vih >= v4lPtr->img->height)) {
		amt[0] = 0;
		amt[1] = 1;
	    } else {
		amt[0] = amt[1] = (double) v4lPtr->yoff / 
		    (double) v4lPtr->img->height;
		amt[1] += (double) v4lPtr->vih / (double) v4lPtr->img->height;
	    }
	    Tcl_PrintDouble(interp, amt[0], buffer);
	    Tcl_AppendElement(interp, buffer);
	    Tcl_PrintDouble(interp, amt[1], buffer);
	    Tcl_AppendElement(interp, buffer);
	} else if (argc == 3) {
	    int yoff;

	    if (Tcl_GetInt(interp, argv[2], &yoff) == TCL_OK) {
		if (v4lPtr->vih < v4lPtr->img->height) {
		    if (yoff < 0) {
			yoff = 0;
		    } else if (yoff + v4lPtr->vih > v4lPtr->img->height) {
			yoff = v4lPtr->img->height - v4lPtr->vih;
		    }
		    v4lPtr->yoff = yoff;
		    ComputeView(v4lPtr);
		    v4lPtr->flags |= REDRAW_NOCLIP | UPDATE_YVIEW;
		    goto redraw;
		}
	    }
	} else if (argc == 4) {
	    double frac;
	    int yoff;

	    if (strcmp(argv[2], "moveto")) {
		goto badyview;
	    }
	    if (Tcl_GetDouble(interp, argv[3], &frac) == TCL_OK) {
		if (v4lPtr->vih < v4lPtr->img->height) {
		    yoff = frac * v4lPtr->img->height;
		    if (yoff < 0) {
			yoff = 0;
		    } else if (yoff + v4lPtr->vih > v4lPtr->img->height) {
			yoff = v4lPtr->img->height - v4lPtr->vih;
		    }
		    v4lPtr->yoff = yoff;
		    ComputeView(v4lPtr);
		    v4lPtr->flags |= REDRAW_NOCLIP | UPDATE_YVIEW;
		    goto redraw;
		}
	    }
	} else if (argc == 5) {
	    int n, isunit = 0;

	    c = argv[4][0];
	    length = strlen(argv[4]);
	    if (strcmp(argv[2], "scroll")) {
		goto badyview;
	    }
	    if ((c == 'u') && (strncmp(argv[4], "units", length) == 0)) {
		isunit = 1;
	    } else if ((c != 'p') || strncmp(argv[4], "pages", length)) {
		goto badyview;
	    }
	    if (Tcl_GetInt(interp, argv[3], &n) == TCL_OK) {
		if (v4lPtr->vih < v4lPtr->img->height) {
		    int yoff;

		    if (isunit) {
			yoff = v4lPtr->yoff + n;
		    } else {
			yoff = v4lPtr->yoff + n * v4lPtr->img->height / 2;
		    }
		    if (yoff < 0) {
			yoff = 0;
		    } else if (yoff + v4lPtr->vih > v4lPtr->img->height) {
			yoff = v4lPtr->img->height - v4lPtr->vih;
		    }
		    v4lPtr->yoff = yoff;
		    ComputeView(v4lPtr);
		    v4lPtr->flags |= REDRAW_NOCLIP | UPDATE_YVIEW;
		    goto redraw;
		}
	    }
	} else {
badyview:
	    Tcl_AppendResult(interp, "bad args to yview subcommand",
			     (char *) NULL);
	    result = TCL_ERROR;
	}


    } else {
	Tcl_AppendResult(interp, "bad option \"", argv[1],
			 "\": must be capture, cget, close, configure, debug, "
			 "devices, fps, geometry, getattr, open, photo, "
			 "record, scan, setattr, xview, or yview",
			 (char *) NULL);
	result = TCL_ERROR;
    }
done:
    Tcl_Release((ClientData) v4lPtr);
    return result;

redraw:
    if (!(v4lPtr->flags & REDRAW_PENDING)) {
	Tcl_DoWhenIdle(DisplayV4lng, (ClientData) v4lPtr);
	v4lPtr->flags |= REDRAW_PENDING;
    }
    Tcl_Release((ClientData) v4lPtr);
    return result;

}

/*
 *----------------------------------------------------------------------
 *
 * DestroyV4lng --
 *
 *	This procedure is invoked by Tcl_EventuallyFree or Tcl_Release
 *	to clean up the internal structure of a v4l at a safe time
 *	(when no-one is using it anymore).
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	Everything associated with the v4l is freed up.
 *
 *----------------------------------------------------------------------
 */

static void
DestroyV4lng(memPtr)
    char *memPtr;		/* Info about v4l widget. */
{
    V4LNG *v4lPtr = (V4LNG *) memPtr;

    CloseVideoDev(v4lPtr);
    if (v4lPtr->islocal) {
	Tk_DeleteGenericHandler(HandleComplEvents, (ClientData) v4lPtr);
	XShmDetach(v4lPtr->display, &v4lPtr->shmInfo);
	shmdt(v4lPtr->shmInfo.shmaddr);
    } else {
	Tcl_Free(v4lPtr->img->data);
    }
    v4lPtr->img->data = NULL;
    XDestroyImage(v4lPtr->img);
    XDestroyRegion(v4lPtr->region);
    if (v4lPtr->iGC != None) {
	XFreeGC(v4lPtr->display, v4lPtr->iGC);
    }
    Tk_FreeOptions(configSpecs, (char *) v4lPtr, v4lPtr->display, 0);
    ng_release_video_buf(&v4lPtr->videoBuf);
#ifdef XV_SUPPORT
    if (v4lPtr->xvFmts != NULL) {
	XFree(v4lPtr->xvFmts);
    }
    if (v4lPtr->xvImg != NULL) {
	XFree(v4lPtr->xvImg);
    }
    if (v4lPtr->xvPort != -1) {
	XvUngrabPort(v4lPtr->display, v4lPtr->xvPort, CurrentTime);
    }
#endif
    Tcl_Free((char *) v4lPtr);
}

/*
 *----------------------------------------------------------------------
 *
 * ConfigureV4lng --
 *
 *	This procedure is called to process an argv/argc list, plus
 *	the Tk option database, in order to configure (or
 *	reconfigure) a v4l widget.
 *
 * Results:
 *	The return value is a standard Tcl result.  If TCL_ERROR is
 *	returned, then interp->result contains an error message.
 *
 * Side effects:
 *	Configuration information, such as text string, colors, font,
 *	etc. get set for v4lPtr;  old resources get freed, if there
 *	were any.
 *
 *----------------------------------------------------------------------
 */

static int
ConfigureV4lng(interp, v4lPtr, argc, argv, flags)
    Tcl_Interp *interp;		/* Used for error reporting. */
    V4LNG *v4lPtr;		/* Information about widget;  may or may
				 * not already have values for some fields. */
    int argc;			/* Number of valid entries in argv. */
    CONST char **argv;		/* Arguments. */
    int flags;			/* Flags to pass to Tk_ConfigureWidget. */
{
    int width, height;

    if (Tk_ConfigureWidget(interp, v4lPtr->tkwin, configSpecs,
			   argc, argv, (char *) v4lPtr, flags) != TCL_OK) {
	return TCL_ERROR;
    }
    if ((v4lPtr->width > 0) || (v4lPtr->height > 0)) {
	width = v4lPtr->width + 
	    2 * (v4lPtr->highlightWidth + v4lPtr->borderWidth);
	height = v4lPtr->height + 
	    2 * (v4lPtr->highlightWidth + v4lPtr->borderWidth);
	Tk_GeometryRequest(v4lPtr->tkwin, width, height);
	ComputeView(v4lPtr);
	v4lPtr->flags |= UPDATE_XVIEW | UPDATE_YVIEW;
    }
    if (Tk_IsMapped(v4lPtr->tkwin)) {
	if (!(v4lPtr->flags & REDRAW_PENDING)) {
	    Tcl_DoWhenIdle(DisplayV4lng, (ClientData) v4lPtr);
	    v4lPtr->flags |= REDRAW_PENDING;
	}
    }
    return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * ComputeView --
 *
 *
 *----------------------------------------------------------------------
 */

static void
ComputeView(v4lPtr)
    V4LNG *v4lPtr;		/* Information about widget. */
{
    int padding = v4lPtr->highlightWidth + v4lPtr->borderWidth;
    int width, height;

    width = Tk_Width(v4lPtr->tkwin) -
	2 * (v4lPtr->highlightWidth + v4lPtr->borderWidth);
    height = Tk_Height(v4lPtr->tkwin) -
	2 * (v4lPtr->highlightWidth + v4lPtr->borderWidth);

    if ((v4lPtr->img->data != NULL) && (v4lPtr->flags & DEV_CAPTURE) &&
	(width > 0) && (height > 0)) {
	int dx, dy, iw, ih;

	dx = dy = padding;
#ifdef XV_SUPPORT
	if (v4lPtr->usexv) {
	    iw = width;
	    ih = height;
	    if (width > v4lPtr->xvMaxWidth) {
		dx = (width - v4lPtr->xvMaxWidth) / 2;
		iw = v4lPtr->xvMaxWidth;
	    }
	    if (height > v4lPtr->xvMaxHeight) {
		dy = (height - v4lPtr->xvMaxHeight) / 2;
		ih = v4lPtr->xvMaxHeight;
	    }
	    v4lPtr->vxoff = dx;
	    v4lPtr->vyoff = dy;
	    v4lPtr->viw = iw;
	    v4lPtr->vih = ih;
	    return;
	}
#endif	
	iw = v4lPtr->img->width;
	ih = v4lPtr->img->height;
	if (width > iw) {
	    dx = (width - iw) / 2;
	}
	if (height > ih) {
	    dy = (height - ih) / 2;
	}
	if (iw > width) {
	    iw = width;
	}
	if (ih > height) {
	    ih = height;
	}
	v4lPtr->vxoff = dx;
	v4lPtr->vyoff = dy;
	v4lPtr->viw = iw;
	v4lPtr->vih = ih;
	if (v4lPtr->viw + v4lPtr->xoff > v4lPtr->img->width) {
	    v4lPtr->xoff = v4lPtr->img->width - v4lPtr->viw;
	}
	if (v4lPtr->vih + v4lPtr->yoff > v4lPtr->img->height) {
	    v4lPtr->yoff = v4lPtr->img->height - v4lPtr->vih;
	}
    } else {
	v4lPtr->vxoff = v4lPtr->vyoff = padding;
	v4lPtr->viw = v4lPtr->vih = 0;
    }
}

/*
 *----------------------------------------------------------------------
 *
 * DisplayV4lng --
 *
 *	This procedure is invoked to display a v4l widget.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	Commands are output to X to display the v4l in its
 *	current mode.
 *
 *----------------------------------------------------------------------
 */

static void
DisplayV4lng(clientData)
    ClientData clientData;	/* Information about widget. */
{
    V4LNG *v4lPtr = (V4LNG *) clientData;
    Tk_Window tkwin = v4lPtr->tkwin;
    int width, height, padding;

    v4lPtr->flags &= ~REDRAW_PENDING;
    if ((tkwin == NULL) || !Tk_IsMapped(tkwin)) {
	v4lPtr->flags &= ~(REDRAW_XEVENTS | REDRAW_NOCLIP);
	XDestroyRegion(v4lPtr->region);
	v4lPtr->region = XCreateRegion();
	return;
    }
    Tcl_Preserve((ClientData) v4lPtr);
    if (v4lPtr->flags & REDRAW_XEVENTS) {
	if (v4lPtr->borderWidth > 0) {
	    Tk_Draw3DRectangle(tkwin, Tk_WindowId(tkwin), v4lPtr->border,
			       v4lPtr->highlightWidth,
			       v4lPtr->highlightWidth,
			       Tk_Width(tkwin) - 2 * v4lPtr->highlightWidth,
			       Tk_Height(tkwin) - 2 * v4lPtr->highlightWidth,
			       v4lPtr->borderWidth, v4lPtr->relief);
	}
	if (v4lPtr->highlightWidth > 0) {
	    XRectangle rects[4];
	    GC gc;

	    padding = 0;
	    width = v4lPtr->highlightWidth;
	    if (v4lPtr->flags & GOT_FOCUS) {
		gc = Tk_GCForColor(v4lPtr->highlightColorPtr,
				   Tk_WindowId(tkwin));
	    } else {
		gc = Tk_GCForColor(v4lPtr->highlightBgColorPtr,
				   Tk_WindowId(tkwin));
	    }
	    rects[0].x = padding;
	    rects[0].y = padding;
	    rects[0].width = Tk_Width(tkwin) - (2 * padding);
	    rects[0].height = width;
	    rects[1].x = padding;
	    rects[1].y = Tk_Height(tkwin) - width - padding;
	    rects[1].width = Tk_Width(tkwin) - (2 * padding);
	    rects[1].height = width;
	    rects[2].x = padding;
	    rects[2].y = width + padding;
	    rects[2].width = width;
	    rects[2].height = Tk_Height(tkwin) - 2*width - 2*padding;
	    rects[3].x = Tk_Width(tkwin) - width - padding;
	    rects[3].y = rects[2].y;
	    rects[3].width = width;
	    rects[3].height = rects[2].height;
	    XFillRectangles(v4lPtr->display, Tk_WindowId(tkwin), gc, rects, 4);
	}
    }

    padding = v4lPtr->highlightWidth + v4lPtr->borderWidth;
    width = Tk_Width(tkwin) - 2 * padding;
    height = Tk_Height(tkwin) - 2 * padding;

    if ((v4lPtr->img->data != NULL) && (v4lPtr->flags & DEV_CAPTURE)) {
	if (v4lPtr->flags & REDRAW_NOCLIP) {
	    XSetClipMask(v4lPtr->display, v4lPtr->iGC, None);
	} else {
	    XSetRegion(v4lPtr->display, v4lPtr->iGC, v4lPtr->region);
	}
	XDestroyRegion(v4lPtr->region);
	v4lPtr->region = XCreateRegion();
	if (v4lPtr->islocal) {
	    if (v4lPtr->flags & COMPL_PENDING) {
		v4lPtr->flags |= REDRAW_AGAIN;
	    } else {
		v4lPtr->flags |= COMPL_PENDING;
#ifdef XV_SUPPORT
		if (v4lPtr->usexv) {
		    XvShmPutImage(v4lPtr->display, v4lPtr->xvPort,
				  Tk_WindowId(tkwin), v4lPtr->iGC,
				  v4lPtr->xvImg, v4lPtr->xoff, v4lPtr->yoff,
				  v4lPtr->img->width, v4lPtr->img->height,
				  v4lPtr->vxoff, v4lPtr->vyoff,
				  v4lPtr->viw, v4lPtr->vih, True);
		} else
#endif
		XShmPutImage(v4lPtr->display,
			     Tk_WindowId(tkwin), v4lPtr->iGC,
			     v4lPtr->img, v4lPtr->xoff, v4lPtr->yoff,
			     v4lPtr->vxoff, v4lPtr->vyoff,
			     v4lPtr->viw, v4lPtr->vih, True);
	    }
	} else {
#ifdef XV_SUPPORT
	    if (v4lPtr->usexv) {
		XvPutImage(v4lPtr->display, v4lPtr->xvPort,
			   Tk_WindowId(tkwin), v4lPtr->iGC,
			   v4lPtr->xvImg, v4lPtr->xoff, v4lPtr->yoff,
			   v4lPtr->img->width, v4lPtr->img->height,
			   v4lPtr->vxoff, v4lPtr->vyoff,
			   v4lPtr->viw, v4lPtr->vih);
	    } else
#endif
	    XPutImage(v4lPtr->display, Tk_WindowId(tkwin), v4lPtr->iGC,
		      v4lPtr->img, v4lPtr->xoff, v4lPtr->yoff,
		      v4lPtr->vxoff, v4lPtr->vyoff,
		      v4lPtr->viw, v4lPtr->vih);
	}
	if (v4lPtr->vxoff > padding) {
	    XFillRectangle(v4lPtr->display, Tk_WindowId(tkwin),
			   v4lPtr->iGC, padding, padding,
			   v4lPtr->vxoff - padding, height);
	}
	if (v4lPtr->vxoff + v4lPtr->viw < width + padding) {
	    int tmp = v4lPtr->vxoff + v4lPtr->viw;

	    XFillRectangle(v4lPtr->display, Tk_WindowId(tkwin),
			   v4lPtr->iGC, tmp, padding,
			   width + padding - tmp, height);
	}
	if (v4lPtr->vyoff > padding) {
	    XFillRectangle(v4lPtr->display, Tk_WindowId(tkwin),
			   v4lPtr->iGC, padding, padding,
			   width, v4lPtr->vyoff - padding);
	}
	if (v4lPtr->vyoff + v4lPtr->vih < height + padding) {
	    int tmp = v4lPtr->vyoff + v4lPtr->vih;

	    XFillRectangle(v4lPtr->display, Tk_WindowId(tkwin),
			   v4lPtr->iGC, padding, tmp,
			   width, height + padding - tmp);
	}
    } else {
	XSetClipMask(v4lPtr->display, v4lPtr->iGC, None);
	if ((width > 0) && (height > 0)) {
	    XFillRectangle(v4lPtr->display, Tk_WindowId(tkwin), v4lPtr->iGC,
			   v4lPtr->highlightWidth + v4lPtr->borderWidth,
			   v4lPtr->highlightWidth + v4lPtr->borderWidth,
			   width, height);
	}
    }
    if ((v4lPtr->flags & UPDATE_XVIEW) && (v4lPtr->xScrollCmd != NULL)) {
	Tcl_Interp *interp = v4lPtr->interp;
	char buffer[128];
	double amt[2];
	int result;

#ifdef XV_SUPPORT
	if (v4lPtr->usexv) {
	    amt[0] = 0;
	    amt[1] = 1;
	} else
#endif
	if ((v4lPtr->viw <= 0) || (v4lPtr->viw >= v4lPtr->img->width)) {
	    amt[0] = 0;
	    amt[1] = 1;
	} else {
	    amt[0] = amt[1] = (double) v4lPtr->xoff /
		(double) v4lPtr->img->width;
	    amt[1] += (double) v4lPtr->viw / (double) v4lPtr->img->width;
	}
	sprintf(buffer, " %g %g", amt[0], amt[1]);
	Tcl_Preserve(interp);
	result = Tcl_VarEval(interp, v4lPtr->xScrollCmd, buffer,
			     (char *) NULL);
	if (result != TCL_OK) {
	    Tcl_AddErrorInfo(interp,
			     "\n    (xscrollcommand executed by v4l)");
	    Tcl_BackgroundError(interp);
	}
	Tcl_Release((ClientData) interp);
    }
    if ((v4lPtr->flags & UPDATE_YVIEW) && (v4lPtr->yScrollCmd != NULL)) {
	Tcl_Interp *interp = v4lPtr->interp;
	char buffer[128];
	double amt[2];
	int result;

#ifdef XV_SUPPORT
	if (v4lPtr->usexv) {
	    amt[0] = 0;
	    amt[1] = 1;
	} else
#endif
	if ((v4lPtr->vih <= 0) || (v4lPtr->vih >= v4lPtr->img->height)) {
	    amt[0] = 0;
	    amt[1] = 1;
	} else {
	    amt[0] = amt[1] = (double) v4lPtr->yoff /
		(double) v4lPtr->img->height;
	    amt[1] += (double) v4lPtr->vih / (double) v4lPtr->img->height;
	}
	sprintf(buffer, " %g %g", amt[0], amt[1]);
	Tcl_Preserve(interp);
	result = Tcl_VarEval(interp, v4lPtr->yScrollCmd, buffer,
			     (char *) NULL);
	if (result != TCL_OK) {
	    Tcl_AddErrorInfo(interp,
			     "\n    (yscrollcommand executed by v4l)");
	    Tcl_BackgroundError(interp);
	}
	Tcl_Release((ClientData) interp);
    }
    v4lPtr->flags &= ~(REDRAW_XEVENTS | REDRAW_NOCLIP |
		       UPDATE_XVIEW | UPDATE_YVIEW);
    Tcl_Release((ClientData) v4lPtr);
}

/*
 *--------------------------------------------------------------
 *
 * V4LNGEventProc --
 *
 *	This procedure is invoked by the Tk dispatcher on
 *	structure changes to a v4l.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	When the window gets deleted, internal structures get
 *	cleaned up.  When it gets exposed, it is redisplayed.
 *
 *--------------------------------------------------------------
 */

static void
V4lngEventProc(clientData, eventPtr)
    ClientData clientData;	/* Information about window. */
    XEvent *eventPtr;		/* Information about event. */
{
    V4LNG *v4lPtr = (V4LNG *) clientData;
    int redraw = 0;

    if (v4lPtr->tkwin == NULL) {
	return;
    }
    switch (eventPtr->type) {
    case Expose: {
	XRectangle rect;

	rect.x = eventPtr->xexpose.x;
	rect.y = eventPtr->xexpose.y;
	rect.width = eventPtr->xexpose.width;
	rect.height = eventPtr->xexpose.height;
	XUnionRectWithRegion(&rect, v4lPtr->region, v4lPtr->region);
	if (eventPtr->xexpose.count == 0) {
	    redraw = 1;
	}
	break;
    }
    case MapNotify:
    case ConfigureNotify:
	v4lPtr->flags |= REDRAW_NOCLIP | UPDATE_XVIEW | UPDATE_YVIEW;
	ComputeView(v4lPtr);
	redraw = 1;
	break;
    case DestroyNotify:
	if (v4lPtr->tkwin != NULL) {
	    Tk_DeleteEventHandler(v4lPtr->tkwin,
				  ExposureMask | StructureNotifyMask |
				  VisibilityChangeMask | PointerMotionMask |
				  ButtonPressMask | ButtonReleaseMask |
				  FocusChangeMask,
				  V4lngEventProc, (ClientData) v4lPtr);
	    v4lPtr->tkwin = NULL;
	    Tcl_DeleteCommandFromToken(v4lPtr->interp, v4lPtr->widgetCmd);
	    v4lPtr->widgetCmd = NULL;
	}
	if (v4lPtr->flags & REDRAW_PENDING) {
	    Tcl_CancelIdleCall(DisplayV4lng, (ClientData) v4lPtr);
	}
	Tcl_EventuallyFree((ClientData) v4lPtr, DestroyV4lng);
	break;
    case FocusIn:
	if (eventPtr->xfocus.detail != NotifyInferior) {
	    if (!(v4lPtr->flags & GOT_FOCUS)) {
		v4lPtr->flags |= GOT_FOCUS;
		redraw = v4lPtr->highlightWidth > 0;
	    }
	}
	break;
    case FocusOut:
	if (eventPtr->xfocus.detail != NotifyInferior) {
	    if (v4lPtr->flags & GOT_FOCUS) {
		v4lPtr->flags &= ~GOT_FOCUS;
		redraw = v4lPtr->highlightWidth > 0;
	    }
	}
	break;
    }
    if (redraw) {
	v4lPtr->flags |= REDRAW_XEVENTS;
	if (!(v4lPtr->flags & REDRAW_PENDING)) {
	    Tcl_DoWhenIdle(DisplayV4lng, (ClientData) v4lPtr);
	    v4lPtr->flags |= REDRAW_PENDING;
	}
    }
}

/*
 *----------------------------------------------------------------------
 *
 * V4lngCmdDeletedProc --
 *
 *	This procedure is invoked when a widget command is deleted.  If
 *	the widget isn't already in the process of being destroyed,
 *	this command destroys it.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	The widget is destroyed.
 *
 *----------------------------------------------------------------------
 */

static void
V4lngCmdDeletedProc(clientData)
    ClientData clientData;	/* Pointer to widget record for widget. */
{
    V4LNG *v4lPtr = (V4LNG *) clientData;
    Tk_Window tkwin = v4lPtr->tkwin;

    /*
     * This procedure could be invoked either because the window was
     * destroyed and the command was then deleted (in which case tkwin
     * is NULL) or because the command was deleted, and then this procedure
     * destroys the widget.
     */

    if (tkwin != NULL) {
	v4lPtr->tkwin = NULL;
	Tk_DestroyWindow(tkwin);
    }
}

/*
 *----------------------------------------------------------------------
 *
 * V4lng_Init --
 *
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	The "v4l" Tcl command is created and the "v4l" package
 *	is registered.
 *
 *----------------------------------------------------------------------
 */

int
V4lng_Init(interp)
    Tcl_Interp *interp;		/* Interpreter in which to create command. */
{
    Tk_Window mainwin;

#ifdef USE_TCL_STUBS
    if (!Tcl_InitStubs(interp, "8.0", 0)) {
	return TCL_ERROR;
    }
#else
    if (Tcl_PkgRequire(interp, "Tk", TK_VERSION, 1) == NULL) {
	return TCL_ERROR;
    }
#endif
#ifdef USE_TK_STUBS
    if (!Tk_InitStubs(interp, "8.0", 0)) {
	return TCL_ERROR;
    }
#else
    if (Tcl_PkgRequire(interp, "Tk", TK_VERSION, 1) == NULL) {
	return TCL_ERROR;
    }
#endif
    ng_init();
    mainwin = Tk_MainWindow(interp);
    if (mainwin == NULL) {
	return TCL_ERROR;
    }
    if (Tcl_PkgProvide(interp, "v4lng", "1.0") != TCL_OK) {
	return TCL_ERROR;
    }
    if (Tcl_CreateCommand(interp, "v4lng", (Tcl_CmdProc *) V4lngCmd,
			  (ClientData) mainwin, (Tcl_CmdDeleteProc *) NULL)
	== NULL) {
	return TCL_ERROR;
    }
    Tcl_GlobalEval(interp, "bind V4LNG <2> {%W scan mark %x %y}\n"
		   "bind V4LNG <B2-Motion> {%W scan dragto %x %y}\n");
    Tcl_GlobalEval(interp, "catch {\n"
		   " bind V4LNG <MouseWheel>"
		   "  {%W yview scroll [expr {- (%D)}] units}\n"
		   " bind V4LNG <Shift-MouseWheel>"
		   "  {%W xview scroll [expr {- (%D)}] units}\n"
		   "}\n");
    Tcl_GlobalEval(interp, "if {\"x11\" eq [tk windowingsystem]} {\n"
		   " bind V4LNG <4> {%W yview scroll -5 units}\n"
		   " bind V4LNG <5> {%W yview scroll 5 units}\n"
		   " bind V4LNG <Shift-Button-4> {%W xview scroll -5 units}\n"
		   " bind V4LNG <Shift-Button-5> {%W xview scroll 5 units}\n"
		   "}\n");
    return TCL_OK;
}
